# %%
import geopandas as gpd

from hydamo_validation import topologic_functions
from hydamo_validation.datamodel import HyDAMO

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

hydamo_version = "2.2"
datamodel = HyDAMO(version=hydamo_version)

dataset_gpkg = DATA_DIR / "tasks" / "test_wrij" / "datasets" / "HyDAMO.gpkg"
stuw_gdf = gpd.read_file(dataset_gpkg, layer="Stuw")
hydroobject_gdf = gpd.read_file(dataset_gpkg, layer="Hydroobject")
duiker_gdf = gpd.read_file(dataset_gpkg, layer="Duikersifonhevel")
gemaal_gdf = gpd.read_file(dataset_gpkg, layer="Gemaal")
aagebied_gdf = gpd.read_file(dataset_gpkg, layer="Afvoergebiedaanvoergebied")

hydroobject_gdf.rename(
    columns={"ruwheidswaardehoog": "ruwheidhoog", "ruwheidswaardelaag": "ruwheidlaag"},
    inplace=True,
)

datamodel.set_data(stuw_gdf, "stuw")
datamodel.set_data(gemaal_gdf, "gemaal")
datamodel.set_data(hydroobject_gdf, "hydroobject")
datamodel.set_data(duiker_gdf, "duikersifonhevel")
datamodel.set_data(aagebied_gdf, "afvoergebiedaanvoergebied")


def test_snaps_to_hydroobject():
    tolerance = 5
    expected_result = [True, True, True, True, True]
    result = topologic_functions.snaps_to_hydroobject(
        datamodel.stuw, datamodel, method="overall", tolerance=tolerance, dtype=bool
    ).to_list()
    assert result == expected_result


def test_length_hydroobject_min():
    expected_result = [
        True,
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
    ]
    result = topologic_functions.geometry_length(
        datamodel.hydroobject.loc[0:10], datamodel=None, length=100, statistic="min"
    ).to_list()
    assert result == expected_result


def test_length_hydroobject_max():
    expected_result = [
        False,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
    ]
    result = topologic_functions.geometry_length(
        datamodel.hydroobject.loc[0:10], datamodel=None, length=100, statistic="max"
    ).to_list()
    assert result == expected_result


def test_not_overlapping():
    expected_result = [True, True, True, True, True, True, True, False, True, True]
    result = topologic_functions.not_overlapping(
        datamodel.duikersifonhevel, datamodel=None, tolerance=0.1
    )[:10].to_list()
    assert result == expected_result


def test_splitted_at_junction():
    expected_result = [True, True, True, True, True, True, True, False, True, True]
    result = topologic_functions.splitted_at_junction(
        datamodel.hydroobject, datamodel=None, tolerance=1
    )[-10:].to_list()
    assert result == expected_result


def test_structures_at_intersections():
    expected_result = [False, True, True, True, True, True, True, True, True, False]
    result = topologic_functions.structures_at_intersections(
        datamodel.hydroobject, datamodel, ["duikersifonhevel"], tolerance=1
    )[-10:].to_list()
    assert result == expected_result


def test_no_dangling_node():
    expected_result = [True, True, False, True, True, True, True, True, False, False]
    result = topologic_functions.no_dangling_node(
        datamodel.hydroobject, datamodel, tolerance=1
    )[-10:].to_list()
    assert result == expected_result


def test_distance_to_others():
    expected_result = [True, False, True, False, True]
    result = topologic_functions.distant_to_others(
        datamodel.stuw, datamodel=None, distance=1
    ).to_list()
    assert result == expected_result


def test_structures_at_boundaries():
    expected_result = [True, True, True, True, False, False, True, True, True, True]
    result = topologic_functions.structures_at_boundaries(
        datamodel.hydroobject,
        datamodel=datamodel,
        areas="afvoergebiedaanvoergebied",
        structures=["gemaal", "stuw", "duikersifonhevel"],
        tolerance=0.01,
        distance=25,
    )[:10].to_list()
    assert result == expected_result


def test_structures_at_nodes():
    expected_result = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        False,
    ]
    result = topologic_functions.structures_at_nodes(
        datamodel.hydroobject, datamodel=datamodel, structures=["stuw"], tolerance=1
    )[:12].to_list()
    assert result == expected_result


# %%



"""Topologic functions executed on extended geodataframe."""

from typing import Literal
import geopandas as gpd
import pandas as pd
from shapely.geometry import Point
from hydamo_validation import geometry

""" 
In this block we define supporting functions. Supporting functions are ignored
by the logical-validation module. They are used by and shared between topological
validation rules.
"""


def _layers_from_datamodel(layers, datamodel):
    if len(layers) > 1:
        series = pd.concat([getattr(datamodel, i)["geometry"] for i in layers])
    else:
        series = getattr(datamodel, layers[0])["geometry"]
    return series.loc[series.index.notna()]


def _lines_snap_at_boundaries(line, other_line, tolerance):
    snaps_start = any(
        line.boundary.geoms[0].distance(i) < tolerance
        for i in other_line.boundary.geoms
    )
    snaps_end = any(
        line.boundary.geoms[-1].distance(i) < tolerance
        for i in other_line.boundary.geoms
    )
    return snaps_start, snaps_end


def _point_not_overlapping_line(point, line, tolerance):
    if line.boundary:
        snaps_start = line.boundary.geoms[0].distance(point) < tolerance
        snaps_end = line.boundary.geoms[-1].distance(point) < tolerance
    else:
        snaps_start = snaps_end = (
            Point(list(line.coords)[0]).distance(point) < tolerance
        )
    if not any([snaps_start, snaps_end]):
        not_overlapping = line.distance(point) > tolerance
    else:
        not_overlapping = True
    return not_overlapping


def _line_not_overlapping_line(line, other_line, tolerance):
    # check if lines snap at boundaries
    snaps_start, snaps_end = _lines_snap_at_boundaries(line, other_line, tolerance)

    # in the case we have two lines with only and-points and all end-points overlap, these are overlapping lines
    if (
        (snaps_start and snaps_end)
        and (len(line.coords) == 2)
        and (len(other_line.coords) == 2)
    ):
        return False

    # compare the shortest line with the longest line
    # if ends snap, we check all coordinates in between, else we check all.
    if line.length < other_line.length:
        not_overlapping = all(
            (
                _point_not_overlapping_line(Point(i), other_line, tolerance)
                for i in line.coords
            )
        )
    else:
        not_overlapping = all(
            (
                _point_not_overlapping_line(Point(i), line, tolerance)
                for i in other_line.coords
            )
        )

    return not_overlapping


def _not_overlapping_line(row, gdf, sindex, tolerance, exclude_row=True):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.buffer(tolerance).bounds))
    if exclude_row:
        indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    if indices:
        gdf_select = gdf.iloc[indices]
        if geometry.geom_type == "LineString":
            not_overlapping = all(
                _line_not_overlapping_line(geometry, i, tolerance)
                for i in gdf_select["geometry"]
            )
        elif geometry.geom_type == "Point":
            not_overlapping = all(
                _point_not_overlapping_line(geometry, i, tolerance)
                for i in gdf_select["geometry"]
            )
    else:
        not_overlapping = True
    return not_overlapping


def _not_overlapping_point(row, gdf, sindex, tolerance, exclude_row=True):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.buffer(tolerance).bounds))
    if exclude_row:
        indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    if indices:
        gdf_select = gdf.iloc[indices]
        not_overlapping = all(
            geometry.distance(i) > tolerance for i in gdf_select["geometry"]
        )
    else:
        not_overlapping = True
    return not_overlapping


def _snap_nodes(row, series, tolerance):
    series_selec = series.loc[~(series.index == row.name)]
    indices = series_selec.loc[
        (series_selec.distance(row["geometry"]) < tolerance)
    ].index.to_list()
    geom = None
    if indices:
        indices.sort()
        if indices[0] < row.name:
            geom = series.loc[indices[0]]
    if geom is None:
        geom = row["geometry"]

    return geom


def _get_nodes(gdf, tolerance):
    # start and end-nodes to GeoSeries
    nodes_series = gdf["geometry"].apply(lambda x: Point(x.coords[0]))
    nodes_series = pd.concat(
        [nodes_series, gdf["geometry"].apply(lambda x: Point(x.coords[-1]))]
    ).reset_index(drop=True)

    # snap nodes within tolerance: nodes within tolerance get the coordinate
    # of the first node.
    nodes_series = gpd.GeoSeries(
        gpd.GeoDataFrame(nodes_series, columns=["geometry"]).apply(
            lambda x: _snap_nodes(x, nodes_series, tolerance), axis=1
        )
    )

    # as all is snapped we can filter unique points
    nodes_series = gpd.GeoSeries(nodes_series.unique())

    return nodes_series


def _only_end_nodes(row, series, sindex, tolerance):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.bounds))
    if indices:
        series_select = series.loc[indices]
        only_end_nodes = all(
            _point_not_overlapping_line(i, geometry, tolerance) for i in series_select
        )
    else:
        only_end_nodes = True

    return only_end_nodes


def _structure_at_intersection(
    geometry, other_geometry, struc_gdf, struc_sindex, tolerance
):
    # compute the intersection and buffer it with tolerance
    intersection = geometry.intersection(other_geometry).buffer(tolerance)

    # get the structures that intersect the intersection bounds
    indices = struc_sindex.intersection(intersection.bounds)

    # check if any of these structures are within intersection
    structure_present = any(i.intersects(intersection) for i in struc_gdf.iloc[indices])

    return structure_present


def _structures_at_intersections(
    row, gdf, sindex, struc_series, struc_sindex, tolerance
):
    geometry = row["geometry"]

    # get a selection of hydroobjects that intersect geometry
    indices = sindex.intersection(geometry.bounds)
    indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    gdf_select = gdf.iloc[indices]
    gdf_select = gdf_select.loc[gdf_select["geometry"].crosses(geometry)]

    # if not emtpy, check if there is a structure at every intersection
    if not gdf_select.empty:
        structure_at_intersections = (
            gdf_select["geometry"]
            .apply(
                lambda x: _structure_at_intersection(
                    x, geometry, struc_series, struc_sindex, tolerance
                )
            )
            .all()
        )
    else:
        structure_at_intersections = True

    return structure_at_intersections


def _intersects_end_node(geometry, series, sindex, tolerance):
    # get inidices in bounds around geometry
    geometry = geometry.buffer(tolerance)
    indices = list(sindex.intersection(geometry.bounds))
    # see if there are start-nodes within within tolerance
    series_select = series.iloc[indices]
    series_select = series_select.loc[series_select.within(geometry)]

    # if dataframe is emtpy, there are no start nodes intersecting end-node
    if series_select.empty:
        intersects_start_node = False
    else:
        intersects_start_node = True
    return intersects_start_node


def _structure_at_boundary(geometry, series, distance):
    return (series.distance(geometry) < distance).any()


def _structures_at_boundaries(
    row, areas_gdf, areas_sindex, struc_series, struc_sindex, tolerance, distance
):
    geometry = row["geometry"]

    # check if hydroobject crosses boundary
    indices = list(areas_sindex.intersection(geometry.bounds))
    areas_gdf_select = areas_gdf.iloc[indices]
    areas_gdf_select = areas_gdf_select.loc[
        areas_gdf_select["geometry"].boundary.intersects(geometry)
    ]

    if areas_gdf_select.empty:
        structure_at_boundary = True
    else:
        # check if hydroobject intersects structures within tolerance
        indices = list(struc_sindex.intersection(geometry.bounds))
        struc_series_select = struc_series.iloc[indices]
        struc_series_select = struc_series_select.loc[
            struc_series_select.distance(geometry) < tolerance
        ]
        if struc_series_select.empty:
            # if there are boundaries, there should be structures
            structure_at_boundary = False
        else:
            # if there are structures they should be close enough to the boundary
            # create series with intersections
            intersections_series = areas_gdf_select["geometry"].boundary.intersection(
                geometry
            )

            # if there is a structure close enough to all boundaries: True. Else: False
            structure_at_boundary = intersections_series.apply(
                lambda x: _structure_at_boundary(x, struc_series_select, distance)
            ).all()

    return structure_at_boundary


def _distant_to_others(row, gdf, sindex, distance):
    geometry = row["geometry"]

    # get a selection of other objects that intersect with
    indices = sindex.intersection(geometry.buffer(distance).bounds)
    indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    gdf_select = gdf.iloc[indices]
    distant_to_others = all(
        i.distance(geometry) > distance for i in gdf_select["geometry"]
    )
    return distant_to_others


def _no_struc_on_line(geometry, struc_series, sindex, tolerance):
    indices = list(sindex.intersection(geometry.bounds))
    if indices:
        struc_series_select = struc_series.iloc[indices]
        no_struc_on_line = all(
            _point_not_overlapping_line(i, geometry, tolerance)
            for i in struc_series_select
        )
    else:
        no_struc_on_line = True
    return no_struc_on_line


def _compare_longitudinal(
    row, parameter, compare_gdf, compare_parameter, direction, logical_operator
):
    branch_id = row["branch_id"]
    select_gdf = compare_gdf.loc[compare_gdf["branch_id"] == branch_id]
    right = None
    result = None

    if not select_gdf.empty:
        if direction == "downstream":
            # select downstream objects
            select_gdf = select_gdf[
                select_gdf["branch_offset"].gt(row["branch_offset"])
            ]
            # if any, select value from closest object
            if not select_gdf.empty:
                right = select_gdf.loc[select_gdf["branch_offset"].idxmin()][
                    compare_parameter
                ]
        elif direction == "upstream":
            # select downstream objects
            select_gdf = select_gdf[
                select_gdf["branch_offset"].lt(row["branch_offset"])
            ]
            # if any, select value from closest object
            if not select_gdf.empty:
                right = select_gdf.loc[select_gdf["branch_offset"].idxmax()][
                    compare_parameter
                ]

    # if there is a right, there is something to compare
    if right is not None:
        if logical_operator == "GT":
            result = row[parameter] > right
        elif logical_operator == "LT":
            result = row[parameter] < right
    return result


""" 
In this block we define topologic functions.
"""


def snaps_to_hydroobject(gdf, datamodel, method, tolerance=0.001, dtype=bool):
    """
    Check if geometries snap to HydroObject

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    method : str, options: 'intersecting', 'overal', 'centroid', 'ends'
        Method that can be used to deterine nearest hydrobject
    tolerance : numeric
        Tolerance used to snap to the hydroobjct
    dtype : dtype, optional
        Dtype to assign to the result series. The default is bool.

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    branches = datamodel.hydroobject
    geometry.find_nearest_branch(
        branches=branches, geometries=gdf, method=method, maxdist=tolerance
    )
    series = ~gdf.branch_offset.isna()
    return series.astype(dtype)


def geometry_length(
    gdf, datamodel, length, statistic: Literal["min", "max"] = "min", dtype=bool
):
    """Check if geometrie length is longer/shorter than a value

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    length : numeric
        Length to compare the geometry length to
    statistic : str, options: 'min', 'max'
        Use length as minimal or maximal length. Default is 'min'
    dtype : dtype, optional
        Dtype to assign to the result series. The default is bool.

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    if statistic == "min":
        series = gdf["geometry"].length >= length
    elif statistic == "max":
        series = gdf["geometry"].length <= length
    return series.astype(dtype)


def not_overlapping(gdf, datamodel, tolerance):
    """Check if an objects LineString geometry is not overlapping other object
    of the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance for overlapping

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    sindex = gdf.sindex
    if (gdf.geom_type == "LineString").all():
        return gdf.apply(
            lambda x: _not_overlapping_line(x, gdf, sindex, tolerance), axis=1
        )
    elif (gdf.geom_type == "Point").all():
        return gdf.apply(
            lambda x: _not_overlapping_point(x, gdf, sindex, tolerance), axis=1
        )
    else:
        raise TypeError(
            f"GeoDataFrame has invalid geometry types: {gdf.geom_type.unique()}. Implemented for this function: [Point, LineString]"
        )


def splitted_at_junction(gdf, datamodel, tolerance):
    """Check if line is splitted when it can considered to be a junction

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance for junction nodes

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    # get the nodes of the hydroobjects within tolerance
    nodes_series = _get_nodes(gdf, tolerance)

    # check for lines if there are nodes on segment outside tolerance of
    # the start-node and end-node.
    sindex = nodes_series.sindex
    return gdf.apply(
        (lambda x: _only_end_nodes(x, nodes_series, sindex, tolerance)), axis=1
    )


def structures_at_intersections(gdf, datamodel, structures, tolerance):
    """Check if there are structures at intersections

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures: str or list
        HyDAMO structures to be expected at intersections
        ("stuw", "duikersifonhevel") Presented as a string or list
    tolerance : numeric
        Max tolerance for junction nodes

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    # make a geodataframe from structures list
    struc_series = _layers_from_datamodel(structures, datamodel)
    # create spatial indices
    sindex = gdf.sindex
    struc_sindex = struc_series.sindex
    # return result
    return gdf.apply(
        lambda x: _structures_at_intersections(
            x, gdf, sindex, struc_series, struc_sindex, tolerance
        ),
        axis=1,
    )


def no_dangling_node(gdf, datamodel, tolerance):
    """Check if the end-node of a linestring object is not within tolerance of
    a start-node of another linestring object in the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance to determine if nodes are connected

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    end_nodes_series = gdf["geometry"].apply(lambda x: Point(x.coords[-1]))
    series = gdf["geometry"].apply(lambda x: Point(x.coords[0]))
    sindex = series.sindex

    return end_nodes_series.apply(
        lambda x: _intersects_end_node(x, series, sindex, tolerance)
    )


def structures_at_boundaries(gdf, datamodel, areas, structures, tolerance, distance):
    """
    Check if there are structures near area (typically water-level areas) boundaries.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    areas : str
        HyDAMO datamodel class with areas ("peilgebiedenpraktijk")
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to dermine if a structure is on the hydroobject
    distance : numeric
        Max distance between structure and area-boundary

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    areas_gdf = getattr(datamodel, areas)
    areas_sindex = areas_gdf.sindex

    struc_series = _layers_from_datamodel(structures, datamodel)
    struc_sindex = struc_series.sindex

    return gdf.apply(
        lambda x: _structures_at_boundaries(
            x, areas_gdf, areas_sindex, struc_series, struc_sindex, tolerance, distance
        ),
        axis=1,
    )


def distant_to_others(gdf, datamodel, distance):
    """
    Check if two objects are sufficient distant to other objects of the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    distance : numeric
        Max distance to other node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    sindex = gdf.sindex

    return gdf.apply(lambda x: _distant_to_others(x, gdf, sindex, distance), axis=1)


def structures_at_nodes(gdf, datamodel, structures, tolerance):
    """
    Check if structures are on boundary-nodes of lines (hydroobjects)

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to determine if a structure is on a node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    struc_series = _layers_from_datamodel(structures, datamodel)
    struc_sindex = struc_series.sindex

    return gdf["geometry"].apply(
        lambda x: _no_struc_on_line(x, struc_series, struc_sindex, tolerance)
    )


def compare_longitudinal(
    gdf,
    datamodel,
    parameter,
    compare_object,
    compare_parameter,
    direction,
    logical_operator,
):
    """
    Check if the value of a parameter in an object is lower/greater than an
    upstream/downstream value of another parameter from another object-layer.


    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to determine if a structure is on a node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    branches = datamodel.hydroobject
    compare_gdf = getattr(datamodel, compare_object)

    # snap layers to to branches
    geometry.find_nearest_branch(
        branches=branches, geometries=compare_gdf, method="overall"
    )

    geometry.find_nearest_branch(branches=branches, geometries=gdf, method="overall")
    return gdf.apply(
        lambda x: _compare_longitudinal(
            x, parameter, compare_gdf, compare_parameter, direction, logical_operator
        ),
        axis=1,
    ).astype(bool)



"""Logical validation."""

# %%
from hydamo_validation import general_functions, logic_functions, topologic_functions
from shapely.geometry import LineString, Point, Polygon
import numpy as np


GEOTYPE_MAPPING = {LineString: "LineString", Point: "Point", Polygon: "Polygon"}
SUMMARY_COLUMNS = [
    "valid",
    "invalid",
    "invalid_critical",
    "invalid_non_critical",
    "ignored",
    "summary",
    "tags_assigned",
    "tags_invalid",
]
LIST_SEPARATOR = ";"
NOTNA_COL_IGNORE = ["related_parameter"]
EXCEPTION_COL = "nen3610id"


def _process_general_function(gdf, function, input_variables):
    return getattr(general_functions, function)(gdf, **input_variables)


def _process_logic_function(gdf, function, input_variables):
    return getattr(logic_functions, function)(gdf, **input_variables)


def _process_topologic_function(gdf, datamodel, function, input_variables):
    return getattr(topologic_functions, function)(gdf, datamodel, **input_variables)


def _notna_indices(gdf, input_variables):
    cols = []
    for k, v in input_variables.items():
        if k not in NOTNA_COL_IGNORE:
            if type(v) is not list:
                v = [v]
            cols += [val for val in v if val in gdf.columns]
    cols = list(set(cols))
    return [index for index, row in gdf[cols].iterrows() if not row.isnull().any()]


def _nan_message(nbr_indices, object_layer, rule_id, rule_type):
    return (
        f"{nbr_indices} objecten niet meegenomen in {rule_type} "
        f"object: '{object_layer}', id: '{rule_id}' wegens no_data in invoer."
    )


def _add_related_gdf(input_variables, datamodel, object_layer):
    related_object = input_variables["related_object"]
    related_gdf = getattr(datamodel, related_object).copy()
    if related_gdf.empty:
        raise Exception(f"Layer '{related_object}' is empty. Rule cannot be executed")

    code_relation = f"{object_layer}id"
    if code_relation not in related_gdf.columns:
        raise KeyError(
            f"'{code_relation}' not in columns of layer '{related_object}': {related_gdf.columns}"
        )
    if "related_parameter" in input_variables.keys():
        related_parameter = input_variables["related_parameter"]
        if related_parameter.startswith("geometry"):
            related_parameter = "geometry"
        if related_parameter not in related_gdf.columns:
            raise KeyError(
                f"'{related_parameter}' not in columns of layer '{related_object}': {related_gdf.columns}"
            )
    input_variables["code_relation"] = code_relation
    input_variables["related_gdf"] = related_gdf
    input_variables.pop("related_object")
    return input_variables


def _add_join_gdf(input_variables, datamodel):
    input_variables["join_gdf"] = getattr(
        datamodel, input_variables["join_object"]
    ).copy()
    return input_variables


def gdf_add_summary(
    gdf,
    variable,
    rule_id,
    penalty,
    error_message,
    critical,
    tags,
    tags_indices,
    separator=LIST_SEPARATOR,
):
    gdf.loc[gdf[variable] == False, "rating"] -= penalty
    gdf.loc[gdf[variable] == False, "summary"] += f"{error_message}{separator}"
    gdf.loc[gdf[variable] == False, "invalid"] += f"{rule_id}{separator}"
    gdf.loc[gdf[variable] == True, "valid"] += f"{rule_id}{separator}"
    gdf.loc[gdf[variable].isna(), "ignored"] += f"{rule_id}{separator}"
    if critical:
        gdf.loc[gdf[variable] == False, "invalid_critical"] += f"{rule_id}{separator}"
    else:
        gdf.loc[gdf[variable] == False, "invalid_non_critical"] += (
            f"{rule_id}{separator}"
        )
    if tags is not None:
        gdf.loc[tags_indices, ("tags_assigned")] += f"{tags}{separator}"
        gdf.loc[gdf[variable] == False, "tags_invalid"] += f"{tags}{separator}"
    return gdf


# %%
def execute(
    datamodel,
    validation_rules_sets,
    layers_summary,
    result_summary,
    logger=None,
    raise_error=False,
):
    """Execute the logical validation."""

    object_rules_sets = [
        i
        for i in validation_rules_sets["objects"]
        if i["object"] in datamodel.data_layers
    ]
    logger.info(
        rf"lagen met valide objecten en regels: {[i['object'] for i in object_rules_sets]}"
    )
    for object_rules in object_rules_sets:
        col_translation: dict = {}

        object_layer = object_rules["object"]
        logger.info(f"{object_layer}: start")
        object_gdf = getattr(datamodel, object_layer).copy()

        # add summary columns
        object_gdf["rating"] = 10
        for col in SUMMARY_COLUMNS:
            object_gdf[col] = ""

        # general rule section
        if "general_rules" in object_rules.keys():
            general_rules = object_rules["general_rules"]
            general_rules_sorted = sorted(general_rules, key=lambda k: k["id"])
            for rule in general_rules_sorted:
                logger.info(
                    f"{object_layer}: uitvoeren general-rule met id {rule['id']}"
                )
                try:
                    result_variable = rule["result_variable"]
                    result_variable_name = (
                        f"general_{rule['id']:03d}_{rule['result_variable']}"
                    )

                    # get function
                    function = next(iter(rule["function"]))
                    input_variables = rule["function"][function]

                    # remove all nan indices
                    indices = _notna_indices(object_gdf, input_variables)
                    dropped_indices = [
                        i
                        for i in object_gdf.index[object_gdf.index.notna()]
                        if i not in indices
                    ]

                    # add object_relation
                    if "related_object" in input_variables.keys():
                        input_variables = _add_related_gdf(
                            input_variables, datamodel, object_layer
                        )
                    elif "join_object" in input_variables.keys():
                        input_variables = _add_join_gdf(input_variables, datamodel)

                    if dropped_indices:
                        result_summary.append_warning(
                            _nan_message(
                                len(dropped_indices),
                                object_layer,
                                rule["id"],
                                "general_rule",
                            )
                        )
                    if object_gdf.loc[indices].empty:
                        object_gdf[result_variable] = np.nan
                    else:
                        result = _process_general_function(
                            object_gdf.loc[indices], function, input_variables
                        )
                        object_gdf.loc[indices, result_variable] = result

                        getattr(datamodel, object_layer).loc[
                            indices, result_variable
                        ] = result

                    col_translation = {
                        **col_translation,
                        result_variable: result_variable_name,
                    }
                except Exception as e:
                    logger.error(
                        f"{object_layer}: general_rule {rule['id']} crashed width Exception {e}"
                    )
                    result_summary.append_error(
                        (
                            "general_rule niet uitgevoerd. Inspecteer de invoer voor deze regel: "
                            f"(object: '{object_layer}', id: '{rule['id']}', function: '{function}', "
                            f"input_variables: {input_variables}, Reason (Exception): {e})"
                        )
                    )
                    if raise_error:
                        raise e
                    else:
                        pass

        validation_rules = object_rules["validation_rules"]
        validation_rules = [
            i for i in validation_rules if ("active" not in i.keys()) | i["active"]
        ]
        validation_rules_sorted = sorted(validation_rules, key=lambda k: k["id"])
        # validation rules section
        for rule in validation_rules_sorted:
            try:
                rule_id = rule["id"]
                logger.info(
                    f"{object_layer}: uitvoeren validatieregel met id {rule_id} ({rule['name']})"
                )
                result_variable = rule["result_variable"]
                if "exceptions" in rule.keys():
                    exceptions = rule["exceptions"]
                    indices = object_gdf.loc[
                        ~object_gdf[EXCEPTION_COL].isin(exceptions)
                    ].index
                else:
                    indices = object_gdf.index
                    exceptions = []
                result_variable_name = (
                    f"validate_{rule_id:03d}_{rule['result_variable']}"
                )

                # get function
                function = next(iter(rule["function"]))
                input_variables = rule["function"][function]

                # remove all nan indices
                notna_indices = _notna_indices(object_gdf, input_variables)
                indices = [i for i in indices[indices.notna()] if i in notna_indices]

                # add object_relation
                if "join_object" in input_variables.keys():
                    input_variables = _add_join_gdf(input_variables, datamodel)

                # apply filter on indices
                if "filter" in rule.keys():
                    filter_function = next(iter(rule["filter"]))
                    filter_input_variables = rule["filter"][filter_function]
                    series = _process_logic_function(
                        object_gdf, filter_function, filter_input_variables
                    )
                    series = series[series.index.notna()]
                    filter_indices = series.loc[series].index.to_list()
                    indices = [i for i in filter_indices if i in indices]
                else:
                    filter_indices = []

                if object_gdf.loc[indices].empty:
                    object_gdf[result_variable] = None
                elif rule["type"] == "logic":
                    object_gdf.loc[indices, (result_variable)] = (
                        _process_logic_function(
                            object_gdf.loc[indices], function, input_variables
                        )
                    )
                elif (rule["type"] == "topologic") and (
                    hasattr(datamodel, "hydroobject")
                ):
                    result_series = _process_topologic_function(
                        # getattr(
                        #     datamodel, object_layer
                        # ),  # FIXME: commented as we need to apply filter in topologic functions as well. Remove after tests pass
                        object_gdf,
                        datamodel,
                        function,
                        input_variables,
                    )
                    object_gdf.loc[indices, (result_variable)] = result_series.loc[
                        indices
                    ]

                col_translation = {
                    **col_translation,
                    result_variable: result_variable_name,
                }

                # summarize
                if rule["error_type"] == "critical":
                    penalty = 5
                    critical = True
                else:
                    penalty = 1
                    critical = False
                if "penalty" in rule.keys():
                    penalty = rule["penalty"]

                error_message = rule["error_message"]

                if "tags" in rule.keys():
                    tags = LIST_SEPARATOR.join(rule["tags"])
                else:
                    tags = None

                exceptions += filter_indices
                _valid_indices = object_gdf[~object_gdf.index.isna()].index
                tags_indices = [i for i in _valid_indices if i not in exceptions]
                object_gdf = gdf_add_summary(
                    gdf=object_gdf,
                    variable=result_variable,
                    rule_id=rule_id,
                    penalty=penalty,
                    error_message=error_message,
                    critical=critical,
                    tags=tags,
                    tags_indices=tags_indices,
                )

            except Exception as e:
                logger.error(
                    f"{object_layer}: validation_rule {rule['id']} width Exception {e}"
                )
                result_summary.append_error(
                    (
                        "validation_rule niet uitgevoerd. Inspecteer de invoer voor deze regel: "
                        f"(object '{object_layer}', rule_id '{rule['id']}', function: '{function}', "
                        f"input_variables: {input_variables}, Reason (Exception): {e})"
                    )
                )
                if raise_error:
                    raise e
                else:
                    pass

        # drop columns
        drop_columns = [
            i
            for i in object_gdf.columns
            if i
            not in list(col_translation.keys())
            + ["nen3610id", "geometry", "rating"]
            + SUMMARY_COLUMNS
        ]
        object_gdf.drop(columns=drop_columns, inplace=True)
        # re_order columns
        column_order = ["nen3610id"]
        column_order += list(col_translation.keys())
        column_order += ["rating"] + SUMMARY_COLUMNS
        if "geometry" in object_gdf.columns:
            column_order += ["geometry"]
        object_gdf = object_gdf[column_order]

        # finish result columns
        for i in SUMMARY_COLUMNS:
            if i in object_gdf.columns:
                object_gdf.loc[:, i] = object_gdf[i].map(lambda x: str(x)[:-1])
        if "rating" in object_gdf.columns:
            object_gdf.loc[:, "rating"] = np.maximum(1, object_gdf["rating"])
        for i in ["tags_assigned", "tags_invalid"]:
            if i in object_gdf.columns:
                object_gdf.loc[:, i] = object_gdf[i].map(
                    lambda x: ";".join(list(set(str(x).split(LIST_SEPARATOR))))
                )

        # rename columns
        object_gdf.rename(columns=col_translation, inplace=True)

        # join gdf to layer_summary
        layers_summary.join_gdf(object_gdf, object_layer)

    return layers_summary, result_summary



from pathlib import Path
from hydamo_validation import validator
import pandas as pd
import pytest
import shutil

DATA_DIR = Path(__file__).parent.joinpath("data")
COVERAGE = {"AHN": DATA_DIR.joinpath(r"dtm")}
DIRECTORY = DATA_DIR.joinpath("tasks", "test_testset")

hydamo_validator = validator(
    output_types=["geopackage"], coverages=COVERAGE, log_level="INFO"
)


@pytest.fixture
def result():
    results_dir = DIRECTORY / "results"
    if results_dir.exists():
        shutil.rmtree(DIRECTORY / "results")
    return hydamo_validator(directory=DIRECTORY, raise_error=True)


def test_results_available(result):
    # assert if result is generated
    assert result is not None


def test_validate_result(result):
    # unpack result
    datamodel, layer_summary, result_summary = result

    # check if profielpunt is available
    assert hasattr(layer_summary, "profielpunt")

    # invalid geometries contain (7) in syntax_geometry column
    result = layer_summary.profielpunt.syntax_geometry.apply(
        lambda x: "(7)" not in x
    ).reset_index(drop=True)

    # we expecte only the first geometry to be valid, containing PointZ, Point, LineString, LineStringZ
    expected_result = pd.Series([True, False, False, False], name="syntax_geometry")
    assert result.equals(expected_result)



# %%
from hydamo_validation import validator
from hydamo_validation import __version__
from pathlib import Path
from datetime import datetime
import shutil

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

coverage = {"AHN": DATA_DIR.joinpath(r"dtm")}
directory = DATA_DIR.joinpath(r"tasks/test_wrij")
exports_dir = Path(__file__).parent / "exports"
if exports_dir.exists():
    shutil.rmtree(exports_dir)
exports_dir.mkdir(exist_ok=True)


hydamo_validator = validator(
    output_types=["geopackage", "csv", "geojson"], coverages=coverage, log_level="INFO"
)


datamodel, layer_summary, result_summary = hydamo_validator(
    directory=directory, raise_error=True
)


def test_non_existing_dir():
    try:
        hydamo_validator(directory="i_do_not_exist", raise_error=True)
        assert False
    except FileNotFoundError:
        assert True


def test_missing_data():
    rules_json = exports_dir.joinpath("validationrules.json")
    if rules_json.exists():
        rules_json.unlink()
    try:
        hydamo_validator(directory=exports_dir, raise_error=True)
        assert False
    except FileNotFoundError:
        assert True


def test_output_type_data():
    _hydamo_validator = validator(
        output_types=["not_supported"], coverages=coverage, log_level="INFO"
    )
    try:
        _hydamo_validator(directory=directory, raise_error=True)
        assert False
    except TypeError:
        assert True


def test_result_success():
    assert result_summary.success


def test_result_version():
    assert result_summary.module_version == __version__


def test_result_check_date():
    try:
        datetime.fromisoformat(result_summary.date_check)
        assert True
    except ValueError:
        assert False


def test_result_duration():
    assert isinstance(result_summary.duration, float)


def test_result_finished():
    assert result_summary.status == "finished"


def test_result_validation_results():
    expected_result = [
        "duikersifonhevel",
        "regelmiddel",
        "kunstwerkopening",
        "stuw",
        "brug",
        "pomp",
        "gemaal",
        "hydroobject",
    ]
    assert all([i in expected_result for i in result_summary.validation_result])


def test_result_dataset_layers():
    expected_result = [
        "Afvoergebiedaanvoergebied",
        "Duikersifonhevel",
        "Hydroobject",
        "Regelmiddel",
        "Stuw",
        "Brug",
        "Bodemval",
        "Gemaal",
        "Kunstwerkopening",
        "Pomp",
    ]
    assert all([i in expected_result for i in result_summary.dataset_layers])


def test_result_result_layers():
    expected_result = [
        "afvoergebiedaanvoergebied",
        "duikersifonhevel",
        "hydroobject",
        "regelmiddel",
        "stuw",
        "brug",
        "bodemval",
        "gemaal",
        "kunstwerkopening",
        "pomp",
    ]
    assert all([i in expected_result for i in result_summary.result_layers])


def test_result_missing_layers():
    expected_result = [
        "admingrenswaterschap",
        "afsluitmiddel",
        "aquaduct",
        "beheergrenswaterschap",
        "bijzonderhydraulischobject",
        "doorstroomopening",
        "grondwaterinfolijn",
        "grondwaterinfopunt",
        "grondwaterkoppellijn",
        "grondwaterkoppelpunt",
        "hydrologischerandvoorwaarde",
        "hydroobject_normgp",
        "lateraleknoop",
        "meetlocatie",
        "meetwaardeactiewaarde",
        "normgeparamprofiel",
        "normgeparamprofielwaarde",
        "peilafwijkinggebied",
        "peilbesluitgebied",
        "peilgebiedpraktijk",
        "peilgebiedvigerend",
        "profielgroep",
        "profiellijn",
        "profielpunt",
        "reglementgrenswaterschap",
        "ruwheidprofiel",
        "streefpeil",
        "sturing",
        "vispassage",
        "vispassagevlak",
        "vuilvang",
        "zandvang",
    ]
    assert all([i in expected_result for i in result_summary.missing_layers])


def test_result_error_layers():
    assert result_summary.error_layers == []


def test_result_syntax_layers():
    expected_result = [
        "afvoergebiedaanvoergebied",
        "duikersifonhevel",
        "hydroobject",
        "regelmiddel",
        "stuw",
        "brug",
        "bodemval",
        "gemaal",
        "kunstwerkopening",
        "pomp",
    ]
    assert all([i in expected_result for i in result_summary.syntax_result])



"""Function to be picked up by the api."""

from typing import List, Callable, Literal, Union
from pathlib import Path
import pandas as pd
from functools import partial
import json
import shutil
import logging
from jsonschema import validate, ValidationError
from json import JSONDecodeError
from hydamo_validation import logical_validation
from hydamo_validation.utils import Timer
from hydamo_validation.summaries import LayersSummary, ResultSummary
from hydamo_validation.datasets import DataSets
from hydamo_validation.datamodel import HyDAMO
from hydamo_validation.syntax_validation import (
    datamodel_layers,
    missing_layers,
    fields_syntax,
)
import traceback

OUTPUT_TYPES = ["geopackage", "geojson", "csv"]
LOG_LEVELS = Literal["INFO", "DEBUG"]
INCLUDE_COLUMNS = ["nen3610id", "code", "categorieoppwaterlichaam"]
SCHEMAS_PATH = Path(__file__).parent.joinpath(r"./schemas")
HYDAMO_SCHEMAS_PATH = SCHEMAS_PATH.joinpath("hydamo")
RULES_SCHEMAS_PATH = SCHEMAS_PATH.joinpath("rules")


def _read_schema(version, schemas_path):
    schema_json = schemas_path.joinpath(rf"rules/rules_{version}.json").resolve()
    with open(schema_json) as src:
        schema = json.load(src)
    return schema


def _init_logger(log_level):
    """Init logger for validator."""
    logger = logging.getLogger(__name__)
    logger.setLevel(getattr(logging, log_level))
    return logger


def _add_log_file(logger, log_file):
    """Add log-file to existing logger"""
    fh = logging.FileHandler(log_file)
    fh.setFormatter(
        logging.Formatter("%(asctime)s %(name)s %(levelname)s - %(message)s")
    )
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    return logger


def _close_log_file(logger):
    """Remove log-file from existing logger"""
    for h in logger.handlers:
        h.close()
        logger.removeHandler(h)


def _log_to_results(log_file, result_summary):
    result_summary.log = log_file.read_text().split("\n")


def read_validation_rules(
    validation_rules_json: Path,
    result_summary: Union[ResultSummary, None] = None,
) -> dict:
    """_summary_

    Parameters
    ----------
    validation_rules_json : Path
        Path to ValidationRules.json()
    result_summary : Union[ResultSummary, None]
        ResultSummary to write exceptions to if specified. Default is None.

    Returns
    -------
    dict
        Validated validationrules

    Raises
    ------
    Exceptions
        - the file with validationrules is not a valid JSON (see exception)
        - schema version cannot be read from validation rules (see exception)
        - validation rules invalid according to json-schema (see exception)
    """
    try:
        validation_rules_sets = json.loads(validation_rules_json.read_text())
    except JSONDecodeError as e:
        if result_summary is not None:
            result_summary.error = [
                "the file with validationrules is not a valid JSON (see exception)"
            ]
        raise e
    try:
        rules_version = validation_rules_sets["schema"]
        schema = _read_schema(rules_version, SCHEMAS_PATH)
    except FileNotFoundError as e:
        if result_summary is not None:
            result_summary.error = [
                "schema version cannot be read from validation rules (see exception)"
            ]
        raise e
    try:
        validate(validation_rules_sets, schema)
    except ValidationError as e:
        if result_summary is not None:
            result_summary.error = [
                "validation rules invalid according to json-schema (see exception)"
            ]
        raise e

    return validation_rules_sets


def validator(
    output_types: List[str] = OUTPUT_TYPES,
    log_level: Literal["INFO", "DEBUG"] = "INFO",
    coverages: dict = {},
) -> Callable:
    """

    Parameters
    ----------
    output_types : List[str], optional
        The types of output files that will be written. Options are
        ["geojson", "csv", "geopackage"]. By default all will be written
    log_level : Literal['INFO', 'DEBUG'], optional
        Level for logger. The default is "INFO".
    coverages : dict, optional
       Location of coverages. E.g. {"AHN: path_to_ahn_dir} The default is {}.

    Returns
    -------
    Callable[[str], dict]
        Partial of _validator function

    """

    return partial(
        _validator,
        output_types=output_types,
        log_level=log_level,
        coverages=coverages,
    )


def _validator(
    directory: str,
    output_types: List[str] = OUTPUT_TYPES,
    log_level: Literal["INFO", "DEBUG"] = "INFO",
    coverages: dict = {},
    raise_error: bool = False,
):
    """
    Parameters
    ----------
    directory : str
        Directory with datasets sub-directory and validation_rules.json
    output_types : List[str], optional
        The types of output files that will be written. Options are
        ["geojson", "csv", "geopackage"]. By default all will be written
    log_level : Literal['INFO', 'DEBUG'], optional
        Level for logger. The default is "INFO".
    coverages : dict, optional
       Location of coverages. E.g. {"AHN: path_to_ahn_dir} The default is {}.
    raise_error: bool, optional
        Will raise an error (or not) when Exception is raised. The default is False

    Returns
    -------
    HyDAMO, LayersSummary, ResultSummary
        Will return a tuple with a filled HyDAMO datamodel, layers_summary and result_summary

    """
    timer = Timer()
    try:
        results_path = None
        dir_path = Path(directory)
        logger = _init_logger(
            log_level=log_level,
        )

        logger.info("init validatie")
        date_check = pd.Timestamp.now().isoformat()
        result_summary = ResultSummary(date_check=date_check)
        layers_summary = LayersSummary(date_check=date_check)
        # check if all files are present
        # create a results_path
        permission_error = False
        if dir_path.exists():
            results_path = dir_path.joinpath("results")
            if results_path.exists():
                try:
                    shutil.rmtree(results_path)
                except PermissionError:
                    permission_error = True
            results_path.mkdir(parents=True, exist_ok=True)
        else:
            raise FileNotFoundError(f"{dir_path.absolute().resolve()} does not exist")

        log_file = results_path.joinpath("validator.log")
        logger = _add_log_file(logger, log_file=log_file)
        logger.info("start validatie")
        if permission_error:
            logger.warning(
                f"Kan pad {results_path} niet verwijderen. Dit kan later tot problemen leiden!"
            )
        dataset_path = dir_path.joinpath("datasets")
        validation_rules_json = dir_path.joinpath("validationrules.json")
        missing_paths = []
        for path in [dataset_path, validation_rules_json]:
            if not path.exists():
                missing_paths += [str(path)]
        if missing_paths:
            result_summary.error += [f"missing_paths: {','.join(missing_paths)}"]
            raise FileNotFoundError(f"missing_paths: {','.join(missing_paths)}")
        else:
            validation_rules_sets = read_validation_rules(
                validation_rules_json, result_summary
            )

        # check if output-files are supported
        unsupported_output_types = [
            item for item in output_types if item not in OUTPUT_TYPES
        ]
        if unsupported_output_types:
            error_message = (
                r"unsupported output types: " f"{','.join(unsupported_output_types)}"
            )
            result_summary.error += [error_message]
            raise TypeError(error_message)

        # set coverages
        if coverages:
            for key, item in coverages.items():
                logical_validation.general_functions._set_coverage(key, item)

        # start validation
        # read data-model
        result_summary.status = "define data-model"
        try:
            hydamo_version = validation_rules_sets["hydamo_version"]
            datamodel = HyDAMO(version=hydamo_version, schemas_path=HYDAMO_SCHEMAS_PATH)
        except Exception as e:
            result_summary.error = ["datamodel cannot be defined (see exception)"]
            raise e

        # validate dataset syntax
        result_summary.status = "syntax-validation (layers)"
        datasets = DataSets(dataset_path)

        result_summary.dataset_layers = datasets.layers

        ## validate syntax of datasets on layers-level and append to result
        logger.info("start syntax-validatie van object-lagen")
        valid_layers = datamodel_layers(datamodel.layers, datasets.layers)
        result_summary.missing_layers = missing_layers(
            datamodel.layers, datasets.layers
        )

        ## validate valid_layers on fields-level and add them to data_model
        result_summary.status = "syntax-validation (fields)"
        syntax_result = []

        ## get status_object if any
        status_object = None
        if "status_object" in validation_rules_sets.keys():
            status_object = validation_rules_sets["status_object"]

        for layer in valid_layers:
            logger.info(f"{layer}: inlezen")

            # read layer
            gdf, schema = datasets.read_layer(
                layer, result_summary=result_summary, status_object=status_object
            )
            if gdf.empty:  # pass if gdf is empty. Most likely due to mall-formed or ill-specifiec status_object
                logger.warning(
                    f"{layer}: geen objecten ingelezen. Zorg dat alle waarden in de kolom 'status_object' voorkomen in {status_object}"
                )
                continue

            layer = layer.lower()
            for col in INCLUDE_COLUMNS:
                if col not in gdf.columns:
                    gdf[col] = None
                    schema["properties"][col] = "str"

            logger.info(f"{layer}: syntax-validatie")
            gdf, result_gdf = fields_syntax(
                gdf,
                schema=schema,
                validation_schema=datamodel.validation_schemas[layer],
                keep_columns=INCLUDE_COLUMNS,
            )

            # Add the syntax-validation result to the results_summary
            layers_summary.set_data(result_gdf, layer, schema["geometry"])
            # Add the corrected datasets_layer data to the datamodel.
            if gdf.empty:
                logger.warning(
                    f"{layer}: geen valide objecten na syntax-validatie. Inspecteer 'syntax_oordeel' in de resultaten; deze is false voor alle objecten. De laag zal genegeerd worden in de (topo)logische validatie."
                )
            else:
                datamodel.set_data(gdf, layer, index_col=None)
            syntax_result += [layer]

        # do logical validation: append result to layers_summary
        result_summary.status = "logical validation"
        logger.info("start (topo)logische validatie van object-lagen")
        layers_summary, result_summary = logical_validation.execute(
            datamodel,
            validation_rules_sets,
            layers_summary,
            result_summary,
            logger,
            raise_error,
        )

        # finish validation and export results
        logger.info("exporteren resultaten")
        result_summary.status = "export results"
        result_layers = layers_summary.export(results_path, output_types)
        result_summary.result_layers = result_layers
        result_summary.error_layers = [
            i for i in datasets.layers if i.lower() not in result_layers
        ]
        result_summary.syntax_result = syntax_result
        result_summary.validation_result = [
            i["object"]
            for i in validation_rules_sets["objects"]
            if i["object"] in result_layers
        ]
        result_summary.success = True
        result_summary.status = "finished"
        result_summary.duration = timer.report()
        logger.info(f"klaar in {result_summary.duration:.2f} seconden")

        _log_to_results(log_file, result_summary)
        result_summary.to_json(results_path)

        _close_log_file(logger)

        return datamodel, layers_summary, result_summary

    except Exception as e:
        stacktrace = rf"\n{traceback.format_exc(limit=0, chain=False)}".split("\n")
        if result_summary.error is not None:
            result_summary.error += stacktrace
        else:
            result_summary.error = stacktrace
        if results_path is not None:
            result_layers = layers_summary.export(results_path, output_types)
            _log_to_results(log_file, result_summary)
            result_summary.to_json(results_path)
        if raise_error:
            raise e
        else:
            result_summary.to_dict()

        _close_log_file(logger)

        return None



from itertools import product
import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import affinity
from shapely.geometry import (
    MultiLineString,
    LineString,
    MultiPolygon,
    Polygon,
    MultiPoint,
    Point,
)


def possibly_intersecting(dataframebounds, geometry, buffer=0):
    """
    Finding intersecting profiles for each branch is a slow process in case of large datasets
    To speed this up, we first determine which profile intersect a square box around the branch
    With the selection, the interseting profiles can be determines much faster.

    Parameters
    ----------
    dataframebounds : numpy.array
    geometry : shapely.geometry.Polygon
    """

    geobounds = geometry.bounds
    idx = (
        (dataframebounds[0] - buffer < geobounds[2])
        & (dataframebounds[2] + buffer > geobounds[0])
        & (dataframebounds[1] - buffer < geobounds[3])
        & (dataframebounds[3] + buffer > geobounds[1])
    )
    # Get intersecting profiles
    return idx


def find_nearest_branch(branches, geometries, method="overall", maxdist=5):
    """
    Method to determine nearest branch for each geometry.
    The nearest branch can be found by finding t from both ends (ends) or the nearest branch from the geometry
    as a whole (overall), the centroid (centroid), or intersecting (intersect).

    Parameters
    ----------
    branches : geopandas.GeoDataFrame
        Geodataframe with branches
    geometries : geopandas.GeoDataFrame
        Geodataframe with geometries to snap
    method='overall' : str
        Method for determine branch
    maxdist=5 : int or float
        Maximum distance for finding nearest geometry
    minoffset : int or float
        Minimum offset from the end of the corresponding branch in case of method=equal
    """
    # Check if method is in allowed methods
    allowed_methods = ["intersecting", "overall", "centroid", "ends"]
    if method not in allowed_methods:
        raise NotImplementedError(f'Method "{method}" not implemented.')

    # Add columns if not present
    if "branch_id" not in geometries.columns:
        geometries["branch_id"] = ""
    if "branch_offset" not in geometries.columns:
        geometries["branch_offset"] = np.nan

    if method == "intersecting":
        # Determine intersection geometries per branch
        geobounds = geometries.bounds.values.T
        for branch in branches.itertuples():
            selectie = geometries.loc[
                possibly_intersecting(geobounds, branch.geometry)
            ].copy()
            intersecting = selectie.loc[selectie.intersects(branch.geometry).values]

            # For each geometrie, determine offset along branch
            for geometry in intersecting.itertuples():
                # Determine distance of profile line along branch
                geometries.at[geometry.Index, "branch_id"] = branch.Index

                # Calculate offset
                branchgeo = branch.geometry
                mindist = min(0.1, branchgeo.length / 2.0)
                offset = round(
                    branchgeo.project(
                        branchgeo.intersection(geometry.geometry).centroid
                    ),
                    3,
                )
                offset = max(mindist, min(branchgeo.length - mindist, offset))
                geometries.at[geometry.Index, "branch_offset"] = offset

    else:
        branch_bounds = branches.bounds.values.T
        # In case of looking for the nearest, it is easier to iteratie over the geometries instead of the branches
        for geometry in geometries.itertuples():
            # Find near branches
            nearidx = possibly_intersecting(
                branch_bounds, geometry.geometry, buffer=maxdist
            )
            selectie = branches.loc[nearidx]

            if method == "overall":
                # Determine distances to branches
                dist = selectie.distance(geometry.geometry)
            elif method == "centroid":
                # Determine distances to branches
                dist = selectie.distance(geometry.geometry.centroid)
            elif method == "ends":
                # Since a culvert can cross a channel, it is
                crds = geometry.geometry.coords[:]
                dist = (
                    selectie["geometry"]
                    .apply(
                        lambda x: max(
                            x.distance(Point(*crds[0])), x.distance(Point(*crds[-1]))
                        )
                    )
                    .astype(float)
                )
                # dist = (
                #     selectie.distance(Point(*crds[0]))
                #     + selectie.distance(Point(*crds[-1]))
                # ) * 0.5

            # Determine nearest
            if dist.min() < maxdist:
                branchidxmin = dist.idxmin()
                geometries.at[geometry.Index, "branch_id"] = dist.idxmin()
                if isinstance(geometry.geometry, Point):
                    geo = geometry.geometry
                else:
                    geo = geometry.geometry.centroid

                # Calculate offset
                branchgeo = branches.at[branchidxmin, "geometry"]
                mindist = min(0.1, branchgeo.length / 2.0)
                offset = max(
                    mindist,
                    min(branchgeo.length - mindist, round(branchgeo.project(geo), 3)),
                )
                geometries.at[geometry.Index, "branch_offset"] = offset



"""functions to be executed on gdf."""

import geopandas as gpd
from typing import Literal
import numpy as np
from pathlib import Path
from rasterstats import zonal_stats
import logging
import pandas as pd

try:
    import rasterio
except ImportError:
    import gdal  # noqa to avoid rasterio.version error: https://github.com/conda-forge/rasterio-feedstock/issues/240
    import rasterio

COVERAGES = {}
# DATA_MODEL = None
# OBJECT_LAYER = None

# We get a false-positive settingwithcopywarning in buffer-function that we supress
pd.options.mode.chained_assignment = None


def _set_coverage(coverage: str, directory: str):
    """Add a coverage for functions."""
    global COVERAGES
    coverage_path = Path(directory)
    if not coverage_path.exists():
        logging.error(
            (
                f"Path to coverage {coverage} does not exist: ",
                f"{coverage_path.absolute().resolve()}",
                ". Functions using this coverage fail without data.",
            )
        )
        raise FileNotFoundError(f"{coverage_path.absolute().resolve()}")
    COVERAGES[coverage] = coverage_path


def _buffer_row(row, column):
    radius = max(row[column], 0.5)
    return row.geometry.buffer(radius)


def _get_geometric_attribute(gdf, geom_parameter):
    geometry, method = geom_parameter.split(".")
    return getattr(gdf[geometry], method)


def sum(gdf, array: list):
    """Return a sum expression."""
    expression = " + ".join(map(str, array))
    return gdf.eval(expression)


def difference(gdf, left, right, absolute=False):
    """
    Difference between 'left' and 'right'

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    absolute : bool, optional
        Absolute (True) or relative difference (False) to left.
        The default is False.

    Returns
    -------
    result : Series
        Float series

    """

    if left in gdf.columns:
        left = gdf[left]
    if right in gdf.columns:
        right = gdf[right]
    if absolute:
        result = (left - right).abs()
    else:
        result = left - right

    return result


def divide(gdf, left, right):
    """
    Division of 'left' by 'right'

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression

    Returns
    -------
    result : Series
        Float series

    """
    expression = " / ".join(map(str, [left, right]))
    return gdf.eval(expression)


def multiply(gdf, left, right):
    """
    Multiply 'left' with 'right'

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : str, numeric
        Right column or value in expression

    Returns
    -------
    result : Series
        Float series

    """
    expression = " * ".join(map(str, [left, right]))
    return gdf.eval(expression)


def buffer(gdf, radius, percentile, coverage="ahn", fill_value: float = None):
    """
    Percentile of coverage-value of an area defined by a radius around the
    object

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    radius: str, numeric
        Radius around object used to define a cirular area
    percentile : int
        The percentile of the coverage within area around object
    coverage : str, optional
        The coverage to use. The default value is 'ahn'
    fill_value : float, optional
        The fill_value to use when the area is not intersecting the coverage.
        The default is None

    Returns
    -------
    result : Series
        Float series

    """
    gdf_out = gdf.copy()
    gdf_out["result"] = np.nan
    xmin, ymin, xmax, ymax = gdf_out.total_bounds
    coverage_path = COVERAGES[coverage]

    index_gdf = gpd.read_file(coverage_path.joinpath("index.shp"))

    for idx, row in index_gdf.cx[xmin:xmax, ymin:ymax].iterrows():
        try:
            bathymetrie_raster = coverage_path.joinpath(
                f'{row["bladnr"].upper()}_CM.tif'
            )

            gdf_select = gdf_out.loc[
                gdf_out["geometry"].centroid.within(row["geometry"])
            ]
            if not gdf_select.empty:
                if isinstance(radius, str):
                    gdf_select.loc[:, ("geometry")] = gdf_select.apply(
                        _buffer_row, args=(radius,), axis=1
                    )
                else:
                    radius = max(radius, 0.5)
                    gdf_select.loc[:, ("geometry")] = gdf_select["geometry"].buffer(
                        radius
                    )

                with rasterio.open(bathymetrie_raster, "r") as src:
                    profile = src.profile
                    raster_data = src.read(1)
                    affine = src.transform
                    scale = src.scales[0]

                raster_stats = zonal_stats(
                    gdf_select,
                    raster_data,
                    affine=affine,
                    stats=f"percentile_{percentile}",
                    nodata=profile["nodata"],
                    raster_out=True,
                )

                gdf_out.loc[gdf_select.index.to_list(), "result"] = [
                    np.nan if item is None else round(item * scale, 2)
                    for item in [
                        item[f"percentile_{percentile}"] for item in raster_stats
                    ]
                ]
        except Exception as e:
            print(
                (
                    f"bathymetrie: {bathymetrie_raster}\n"
                    f"indices: {gdf_select.index}\n"
                    f"geometrien: {gdf_select['geometry']}"
                )
            )
            raise e

    # fill series if if provided
    if fill_value is not None:
        gdf_out.loc[gdf_out["result"].isna(), "result"] = fill_value

    return gdf_out["result"]


def join_parameter(
    gdf,
    join_object: str,
    join_gdf: gpd.GeoDataFrame,
    join_parameter: str,
    fill_value=None,
):
    """Joins a parameteer of other object to geodataframe."""
    _gdf = gdf.copy()

    _join_gdf = join_gdf.copy()
    _join_gdf.set_index("globalid", inplace=True)
    series = _join_gdf[join_parameter]
    series.name = "result"
    _gdf = _gdf.merge(series, how="left", left_on=f"{join_object}id", right_index=True)

    # fill series if if provided
    if fill_value is not None:
        _gdf.loc[_gdf["result"].isna(), "result"] = fill_value

    return _gdf["result"]


def object_relation(
    gdf,
    related_gdf: gpd.GeoDataFrame,
    code_relation: str,
    statistic: Literal["min", "max", "sum", "count"],
    related_parameter: str = None,
    fill_value=None,
):
    """
    Statistic of related object to geodataframe

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    related_gdf : GeoDataFrame
        GeoDataFrame with related attributes
    code_relation : str
        Column in related_gdf used to relate to gdf. Example 'stuwid'
    statistic : str, options: 'min', 'max', 'sum', 'count'
        Statistic to compute over related values
    related_parameter: str
        Column in related_gdf over which the statistic is to be computed
    fill_value : float, optional
        The fill_value to use when the area is not intersecting the coverage.
        The default is None

    Returns
    -------
    result : Series
        Float series

    """

    gdf_out = gdf.copy()

    # remove NaN values in from related_gdf[related_parameter]
    if related_parameter:
        if "geometry" in related_parameter:
            related_gdf[related_parameter] = _get_geometric_attribute(
                related_gdf, related_parameter
            )
        related_gdf = related_gdf.loc[related_gdf[related_parameter].notna()]

    # compute statistic
    if statistic == "count":
        series = related_gdf.groupby(by=[code_relation])[code_relation].count()
    elif statistic == "sum":
        series = related_gdf.groupby(by=[code_relation])[related_parameter].sum()
    elif statistic == "min":
        series = related_gdf.groupby(by=[code_relation])[related_parameter].min()
    elif statistic == "max":
        series = related_gdf.groupby(by=[code_relation])[related_parameter].max()
    elif statistic == "majority":
        series = related_gdf.groupby(by=[code_relation])[related_parameter].agg(
            pd.Series.mode
        )

    # join series with gdf
    series.name = "result"
    series = pd.DataFrame(series.loc[series.index.isin(gdf["globalid"])]).reset_index()
    gdf_out = gdf_out.merge(
        series, how="left", left_on="globalid", right_on=code_relation
    )

    # fill series if if provided
    if fill_value is not None:
        gdf_out.loc[gdf_out["result"].isna(), "result"] = fill_value
    return gdf_out["result"]



# %%
from hydamo_validation.summaries import LayersSummary, ResultSummary, OUTPUT_TYPES
from hydamo_validation.datasets import DataSets
from pathlib import Path
import shutil

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR


dataset_path = DATA_DIR.joinpath(r"tasks/test_wrij/datasets")
exports_dir = Path(__file__).parent / "exports"
exports_dir.mkdir(exist_ok=True)

datasets = DataSets(dataset_path)
layers_summary = LayersSummary()
result_summary = ResultSummary()
result_summary.dataset_layers = datasets.layers

gdf, _ = datasets.read_layer("Hydroobject")

gdf.drop(
    [i for i in gdf.columns if i not in ["nen3610id", "geometry"]], axis=1, inplace=True
)


def test_layers_set_data():
    _gdf = gdf.copy()
    _gdf["syntax_1"] = ""
    _gdf["syntax_oordeel"] = True
    layers_summary.set_data(_gdf, "hydroobject", geo_type="LineString")
    assert layers_summary.hydroobject.equals(_gdf)


def test_layers_join_gdf():
    _gdf = gdf.copy()
    _gdf["general_1"] = 0.1
    _gdf["validate_1"] = True
    _gdf["rating"] = 10
    _gdf["tags"] = ""
    layers_summary.join_gdf(_gdf, "hydroobject")
    expected_columns = [
        "syntax_1",
        "syntax_oordeel",
        "nen3610id",
        "general_1",
        "validate_1",
        "rating",
        "tags",
        "geometry",
    ]
    assert all([i in expected_columns for i in layers_summary.hydroobject.columns])


def test_layers_export():
    result = exports_dir / "results"
    if result.exists():
        shutil.rmtree(result)
    result.mkdir(exist_ok=True)
    layers_summary.export(result, output_types=OUTPUT_TYPES)
    result_gpkg = result / "results.gpkg"
    if result_gpkg.exists():
        shutil.rmtree(result)
        assert True
    else:
        shutil.rmtree(result)
        assert False


def test_result_to_json():
    result_json = exports_dir / "validation_result.json"
    if result_json.exists():
        result_json.unlink()
    result_summary.to_json(exports_dir)
    if result_json.exists():
        assert True
        result_json.unlink()
    else:
        assert False
        result_json.unlink()


def test_result_to_dict():
    result = result_summary.to_dict()
    assert result["dataset_layers"] == datasets.layers


def test_result_to_all():
    result_json = exports_dir / "validation_result.json"
    if result_json.exists():
        result_json.unlink()
    result = result_summary.to_all(exports_dir)
    if (result_json.exists()) & (result["dataset_layers"] == datasets.layers):
        assert True
        result_json.unlink()
    else:
        assert False
        result_json.unlink()


def test_append_warning():
    result_summary.append_warning("my warning")
    assert result_summary.warnings == ["my warning"]


def test_append_error():
    result_summary.append_error("my error")
    assert result_summary.errors == ["my error"]



__author__ = ["Het Waterschapshuis", "D2HYDRO", "HKV", "HydroConsult"]
__copyright__ = "Copyright 2021, HyDAMO ValidatieTool"
__credits__ = ["D2HYDRO", "HKV", "HydroConsult"]
__version__ = "1.4.1b1"

__license__ = "MIT"
__maintainer__ = "Daniel Tollenaar"
__email__ = "daniel@d2hydro.nl"

import fiona  # top-level import to avoid fiona import issue: https://github.com/conda-forge/fiona-feedstock/issues/213
from pathlib import Path
from hydamo_validation.functions import topologic as topologic_functions
from hydamo_validation.functions import logic as logic_functions
from hydamo_validation.functions import general as general_functions
from hydamo_validation.validator import validator

__all__ = [
    "fiona",
    "topologic_functions",
    "logic_functions",
    "general_functions",
    "validator",
    "SCHEMAS_PATH",
]



from hydamo_validation.validator import read_validation_rules
from jsonschema import ValidationError
from json import JSONDecodeError

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR


def test_schema_1_1():
    validation_rules_json = DATA_DIR.joinpath(
        "validation_rules", "ValidationRules_1.1.json"
    )

    validation_rules = read_validation_rules(validation_rules_json)
    assert validation_rules["schema"] == "1.1"


def test_schema_1_2():
    validation_rules_json = DATA_DIR.joinpath(
        "validation_rules", "ValidationRules_1.2.json"
    )

    validation_rules = read_validation_rules(validation_rules_json)
    assert validation_rules["schema"] == "1.2"


def test_invalid_json():
    validation_rules_json = DATA_DIR.joinpath("validation_rules", "NotAjson.json")

    try:
        read_validation_rules(validation_rules_json)
        assert False
    except JSONDecodeError:
        assert True


def test_missing_schema():
    validation_rules_json = DATA_DIR.joinpath(
        "validation_rules", "MissingSchemaValidationRules_1.1.json"
    )

    try:
        read_validation_rules(validation_rules_json)
        assert False
    except FileNotFoundError:
        assert True


def test_validation_error():
    validation_rules_json = DATA_DIR.joinpath(
        "validation_rules", "MissingOrderValidationRules_1.1.json"
    )

    try:
        read_validation_rules(validation_rules_json)
        assert False
    except ValidationError:
        assert True



from hydamo_validation.datasets import DataSets

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR


dataset_path = DATA_DIR.joinpath(r"tasks/test_wrij/datasets")
datasets = DataSets(dataset_path)


def test_layers():
    expected_result = [
        "Afvoergebiedaanvoergebied",
        "Duikersifonhevel",
        "Hydroobject",
        "Regelmiddel",
        "Stuw",
        "Brug",
        "Bodemval",
        "Gemaal",
        "Kunstwerkopening",
        "Pomp",
    ]
    assert datasets.layers == expected_result


def test_read_layer():
    gdf, schema = datasets.read_layer(layer="Duikersifonhevel")
    assert len(gdf) == 55


def test_read_layer_filter_status():
    gdf, schema = datasets.read_layer(
        layer="Duikersifonhevel", status_object=["planvorming", "gerealiseerd"]
    )
    assert len(gdf) == 53


def test_read_layer_not_available():
    try:
        datasets.read_layer(layer="hydroobject")
        assert False
    except KeyError:
        assert True



# %%
from hydamo_validation import validator
from pathlib import Path
import pandas as pd
import shutil

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR
coverage = {"AHN": DATA_DIR.joinpath(r"dtm")}
directory = DATA_DIR.joinpath(r"tasks/test_dommelerwaard")
exports_dir = Path(__file__).parent / "exports"
if exports_dir.exists():
    shutil.rmtree(exports_dir)
exports_dir.mkdir(exist_ok=True)

hydamo_validator = validator(
    output_types=["geopackage", "csv", "geojson"], coverages=coverage, log_level="INFO"
)


datamodel, layer_summary, result_summary = hydamo_validator(
    directory=directory, raise_error=False
)


def test_hydroobjects():
    EXPECTED_INDEX = pd.Index(
        [
            "nen3610id",
            "code",
            "categorieoppwaterlichaam",
            "geometry",
            "syntax_breedte",
            "syntax_categorieoppwaterlichaam",
            "syntax_code",
            "syntax_created_date",
            "syntax_detailniveaugeometrie",
            "syntax_draineert",
            "syntax_featuretype",
            "syntax_getijdebeinvloed",
            "syntax_globalid",
            "syntax_hydroobjectgroeperingid",
            "syntax_hyperlink",
            "syntax_krwstromendid",
            "syntax_last_edited_date",
            "syntax_lengte",
            "syntax_lvpublicatiedatum",
            "syntax_metadataid",
            "syntax_naam",
            "syntax_namespace",
            "syntax_nen3610id",
            "syntax_niveau",
            "syntax_objectid",
            "syntax_ontstaanswijze",
            "syntax_opmerking",
            "syntax_oppervlaktewaterlichaamid",
            "syntax_oppwatervolgnummer",
            "syntax_persistentie",
            "syntax_ruimtelijkeafbakeningbekend",
            "syntax_ruwheidhoog",
            "syntax_ruwheidlaag",
            "syntax_soortoppwaterkwaliteit",
            "syntax_soortoppwaterkwantiteit",
            "syntax_statusleggerwatersysteem",
            "syntax_statusobject",
            "syntax_typeruwheid",
            "syntax_vaarwegid",
            "syntax_geometry",
            "syntax_oordeel",
            "validate_000_lengte_gt_1m",
            "validate_001_goed_gesplitst",
            "validate_002_goede_kruising",
            "validate_003_geen_eindpunt",
            "validate_006_kunstwerk_op_peilscheiding",
            "validate_005_kunstwerk_tussen_hobject",
            "rating",
            "valid",
            "invalid",
            "invalid_critical",
            "invalid_non_critical",
            "ignored",
            "summary",
            "tags_assigned",
            "tags_invalid",
            "date_check",
        ]
    )

    assert layer_summary.hydroobject.columns.equals(EXPECTED_INDEX)


# %%
